from pathlib import Path
from typing import Sequence, Tuple, Union

import torch
from torch import Tensor

import torchjpeg.codec
from torchjpeg.dct import Stats

from .image_list import ImageList
from .jpeg_quantized_dataset import _prep_coefficients


class FolderOfJpegDataset(torch.utils.data.Dataset):
    """
    Loads coefficents from a folder of JPEG without any labels. For each image, it returns the format of :py:func:`torchjpeg.codec.read_coefficients`. The
    images must be actualy JPEG files (stored as JPEGs) for this to work. The relative path to the JPEG file will be returned along with the coefficients. The
    coefficients themselves are not guaranteed to be the same size, use the collate function to collate these into a batched Tensor by adding padding.

    Args:
        path (Path): The path to load images from
        stats (Stats): DCT stats to use to normalize the coefficients
        extensions (List[str]): The JPEG file extensions to search for
    """

    def __init__(self, path: Union[str, Path], stats: Stats, extensions: Sequence[str] = [".jpg", ".jpeg", ".JPEG"]):
        # pylint: disable=dangerous-default-value
        if isinstance(path, str):
            path = Path(path)

        self.path = path
        self.stats = stats

        if path.is_dir():
            self.images = list(filter(lambda p: p.suffix in extensions, path.glob("**/*")))
        else:
            self.images = [path]

    def __len__(self):
        return len(self.images)

    def __getitem__(self, idx):
        image = self.images[idx]

        dim, quantization, Y_coefficients, CbCr_coefficients = torchjpeg.codec.read_coefficients(str(image))  # type: ignore

        y_dequantized, cbcr_dequantized, y_q, c_q = _prep_coefficients(quantization, Y_coefficients, CbCr_coefficients, self.stats)

        return y_dequantized.squeeze(0), cbcr_dequantized.squeeze(0), y_q.unsqueeze(0), c_q.unsqueeze(0), image.relative_to(self.path), dim[0]

    @staticmethod
    def collate(batch_list: Sequence[Tuple[Tensor, Tensor, Tensor, Tensor, Path, Tensor]]) -> Tuple[Tensor, Tensor, Tensor, Tensor, Sequence[Path], Tensor]:
        """
        Custom collate function which works for return values from this dataset. Adds padding to the images so that they can be stored in a single tensor

        Args:
            batch_list: Output from this dataset

        Returns:
            Batch with each input collated into single tensors
        """
        y_coefs = []
        cbcr_coefs = []

        yqs = torch.stack([b[2] for b in batch_list])
        cqs = torch.stack([b[3] for b in batch_list])
        sizes = torch.stack([b[5] for b in batch_list])

        paths = [b[4] for b in batch_list]

        for b in batch_list:
            y_coefs.append(b[0])
            cbcr_coefs.append(b[1])

        y_coefs_t = ImageList.from_tensors(y_coefs).tensor
        cbcr_coefs_t = ImageList.from_tensors(cbcr_coefs).tensor

        return y_coefs_t, cbcr_coefs_t, yqs, cqs, paths, sizes



# pylint: disable=missing-function-docstring
import pathlib
from subprocess import check_call

from torch.utils.cpp_extension import BuildExtension, CppExtension


def build_libjpeg():
    check_call("cd src/libjpeg && ./configure --enable-static --with-pic && make", shell=True)


def build(setup_kwargs):
    build_libjpeg()

    libjpeg_dir = pathlib.Path(__file__).parent.absolute() / "src" / "libjpeg"

    setup_kwargs.update(
        {
            "ext_modules": [
                CppExtension(
                    "torchjpeg.codec._codec_ops",
                    [
                        "src/torchjpeg/codec/codec_ops.cpp",
                    ],
                    include_dirs=[str(libjpeg_dir)],
                    extra_objects=[str(libjpeg_dir / ".libs" / "libjpeg.a")],
                    extra_compile_args=["-std=c++17"],
                )
            ],
            "cmdclass": {"build_ext": BuildExtension},
        }
    )


if __name__ == "__main__":
    build_libjpeg()



r"""
:py:mod:`torchjpeg.quantization.ijg` provides functions which match the Independent JPEG Group's libjpeg quantization method.
"""
import torch
from torch import Tensor

from torchjpeg.dct import batch_dct, batch_idct

from ._quantize import dequantize, quantize

# Don't mess with matrix formatting
# fmt: off
luma_quant_matrix = torch.tensor([
    16,  11,  10,  16,  24,  40,  51,  61,
    12,  12,  14,  19,  26,  58,  60,  55,
    14,  13,  16,  24,  40,  57,  69,  56,
    14,  17,  22,  29,  51,  87,  80,  62,
    18,  22,  37,  56,  68, 109, 103,  77,
    24,  35,  55,  64,  81, 104, 113,  92,
    49,  64,  78,  87, 103, 121, 120, 101,
    72,  92,  95,  98, 112, 100, 103,  99
])
# fmt: on

# Don't mess with matrix formatting
# fmt: off
chroma_quant_matrix = torch.tensor([
    17,  18,  24,  47,  99,  99,  99,  99,
    18,  21,  26,  66,  99,  99,  99,  99,
    24,  26,  56,  99,  99,  99,  99,  99,
    47,  66,  99,  99,  99,  99,  99,  99,
    99,  99,  99,  99,  99,  99,  99,  99,
    99,  99,  99,  99,  99,  99,  99,  99,
    99,  99,  99,  99,  99,  99,  99,  99,
    99,  99,  99,  99,  99,  99,  99,  99
])
# fmt: on

quantization_max = 255.0


def qualities_to_scale_factors(qualities: Tensor) -> Tensor:
    r"""
    Converts a batch of qualities in [0, 100] to a batch of scale factors suitable for scaling one of the IJG reference quantization matrices.

    Args:
        qualities (Tensor): A single dimensional batch of qualities.

    Returns:
        Tensor: A single dimensional batch of scale factors.
    """
    qualities = qualities.clone()
    qualities[qualities <= 0] = 1
    qualities[qualities > 100] = 100

    indices_0_50 = qualities < 50
    indices_50_100 = qualities >= 50

    qualities[indices_0_50] = 5000 // qualities[indices_0_50]
    qualities[indices_50_100] = torch.trunc(200 - qualities[indices_50_100] * 2)

    return qualities


def scale_quantization_matrices(scale_factor: Tensor, table: str = "luma") -> Tensor:
    r"""
    Scales one of the IJG reference quantization matrices.

    Args:
        scale_factor (Tensor): A batch of :math:`N` scale factors.
        table (str): A string indicating the table to use, either "luma" or "chroma"

    Returns:
        Tensor: A batch of quantization matrices of shape :math:`(N, 64)`.
    """
    if table == "luma":
        t = luma_quant_matrix
    elif table == "chroma":
        t = chroma_quant_matrix

    t = t.to(scale_factor.device)

    t = t.unsqueeze(0)
    scale_factor = scale_factor.unsqueeze(1)

    mat = (t * scale_factor + 50) // 100
    mat[mat <= 0] = 1
    mat[mat > 255] = 255

    return mat


def get_coefficients_for_qualities(quality: Tensor, table: str = "luma") -> Tensor:
    r"""
    Gets IJG quantization matrices for a given batch of qualities.

    Args:
        quality (Tensor): A batch of qualities of shape :math:`(N)`
        table (str): A string indicating the table to use, either "luma" or "chroma"

    """
    scaler = qualities_to_scale_factors(quality)
    mat = scale_quantization_matrices(scaler, table=table)
    return mat.view(-1, 1, 8, 8)


def quantize_at_quality(dct: Tensor, quality: int, table: str = "luma") -> Tensor:
    r"""
    Quantizes using a scalar quality instead of a quantization matrix. Uses IJG quantization matrices.

    Args:
        dct (Tensor): DCT coefficients of shape :math:`(N, 1, H ,W)`.
        quality (int): A scalar in [0, 100] specifying the desired quality.
        table (str): One of "luma" or "chroma" to choose the desired set of tables.

    Returns:
        Tensor: Quantized DCT coefficients.
    """
    mat = get_coefficients_for_qualities(torch.tensor([quality]), table=table)
    return quantize(dct, mat)


def dequantize_at_quality(dct: Tensor, quality: int, table: str = "luma") -> Tensor:
    r"""
    Dequantizes using a scalar quality instead of a quantization matrix. uses IJG quantization matrices.

    Args:
        dct (Tensor): Quantized DCT coefficients of shape :math:`(N, 1, H, W)`.
        quality (int): A scalar in [0, 100] specifying the quality that the coefficients were quantized at.
        table (str): One of "luma" or "chroma" to choose the desired set of tables.
    """
    mat = get_coefficients_for_qualities(torch.tensor([quality]), table=table)
    return dequantize(dct, mat)


def compress_coefficients(batch: Tensor, quality: int, table: str = "luma") -> Tensor:
    r"""
    A high level function that takes a batch of pixels in [0, 1] and returns quantized DCT coefficients.

    Args:
        batch (Tensor): A batch of images to quantize of shape `(N, 1, H, W)`.
        quality (int): A scalar quality in [0, 100] specifying the desired quality.
        table (str): One of "luma" or "chroma" to choose the desired set of tables.

    Returns:
        Tensor: A batch of quantized DCT coefficients.
    """
    batch = batch * 255 - 128
    d = batch_dct(batch)
    d = quantize_at_quality(d, quality, table=table)
    return d


def decompress_coefficients(batch: Tensor, quality: int, table: str = "luma") -> Tensor:
    r"""
    A high level function that converts quantized DCT coefficients to pixels.

    Args:
        batch (Tensor): A batch of quantized DCT coefficients of shape `(N, 1, H, W)`.
        quality (int): A scalar quality in [0, 100] specifying the quality the coefficients were quantized at.
        table (str): One of "luma" or "chroma" to choose the desired set of tables.

    Returns:
        Tensor: A batch of image pixels.
    """
    d = dequantize_at_quality(batch, quality, table=table)
    d = batch_idct(d)
    d = (d + 128) / 255
    return d



import torch
from torch import Tensor


def ssim_single(
    image: Tensor,
    target: Tensor,
) -> Tensor:
    r"""
    Computes SSIM for a single channel
    """
    C1 = 0.01**2
    C2 = 0.03**2

    avg_filter = torch.ones(1, 1, 8, 8, device=image.device) / 64

    mu_i = torch.nn.functional.conv2d(image, avg_filter)
    mu_t = torch.nn.functional.conv2d(target, avg_filter)

    var_i = torch.nn.functional.conv2d(image**2, avg_filter) - mu_i**2
    var_t = torch.nn.functional.conv2d(target**2, avg_filter) - mu_t**2
    cov_it = torch.nn.functional.conv2d(target * image, avg_filter) - mu_i * mu_t

    ssim_blocks = ((2 * mu_i * mu_t + C1) * (2 * cov_it + C2)) / ((mu_i**2 + mu_t**2 + C1) * (var_i + var_t + C2))
    return ssim_blocks.view(image.shape[0], -1).mean(1)


def ssim(image: Tensor, target: Tensor) -> Tensor:
    r"""
    Computes the structural similarity index of two images as defined in [1].

    Args:
        image (Tensor): The input images of shape :math:`(N, C, H, W)`.
        target (Tensor): The target images of shape :math:`(N, C, H, W)`.

    Returns:
        Tensor: The SSIM of each image of shape :math:`(N)`.

    Note:
        This function uses an :math:`8 \times 8` uniform averaging window used in JPEG evaluation tasks instead of the :math:`11 \times 11` gaussian window
        used in the original paper and by default in other SSIM implementations.

        [1] Wang, Zhou, et al. "Image quality assessment: from error visibility to structural similarity." IEEE transactions on image processing 13.4 (2004): 600-612.
    """
    total = torch.stack([ssim_single(image[:, c : c + 1, :, :], target[:, c : c + 1, :, :]) for c in range(target.shape[1])]).sum(0)
    return total / target.shape[1]



r"""
The :code:`torchjpeg.dct` package provides utilities for performing forward and inverse discrete cosine transforms on images.
The dct routines are implemented in pytorch so they can be GPU accelerated and differentiated.
While the routines here are restricted to two dimensional signals, the block size is configurable e.g. the DCT does
not need to be performed on only the :math:`8 \times 8` block size used by JPEG.
This package includes additional utilities for splitting images into non-overlapping blocks, performing fast
color transforms on Tensors, and normalizing DCT coefficients as preparation for input to a CNN.
"""
import math
from typing import Optional, Sequence, Tuple

import torch
from torch import Tensor
from torch.nn.functional import pad

from ._block import blockify, deblockify
from ._color import to_rgb, to_ycbcr
from ._nn import double_nn_dct, half_nn_dct
from ._stats import Stats


def _normalize(N: int) -> Tensor:
    r"""
    Computes the constant scale factor which makes the DCT orthonormal
    """
    n = torch.ones((N, 1))
    n[0, 0] = 1 / math.sqrt(2)
    return n @ n.t()


def _harmonics(N: int) -> Tensor:
    r"""
    Computes the cosine harmonics for the DCT transform
    """
    spatial = torch.arange(float(N)).reshape((N, 1))
    spectral = torch.arange(float(N)).reshape((1, N))

    spatial = 2 * spatial + 1
    spectral = (spectral * math.pi) / (2 * N)

    return torch.cos(spatial @ spectral)


def block_dct(blocks: Tensor) -> Tensor:
    r"""
    Computes the DCT of image blocks

    Args:
        blocks (Tensor): Non-overlapping blocks to perform the DCT on in :math:`(N, C, L, H, W)` format.

    Returns:
        Tensor: The DCT coefficients of each block in the same shape as the input.

    Note:
        The function computes the forward DCT on each block given by

        .. math::

            D_{i,j}={\frac {1}{\sqrt{2N}}}\alpha (i)\alpha (j)\sum _{x=0}^{N}\sum _{y=0}^{N}I_{x,y}\cos \left[{\frac {(2x+1)i\pi }{2N}}\right]\cos \left[{\frac {(2y+1)j\pi }{2N}}\right]

        Where :math:`i,j` are the spatial frequency indices, :math:`N` is the block size and :math:`I` is the image with pixel positions :math:`x, y`.

        :math:`\alpha` is a scale factor which ensures the transform is orthonormal given by

        .. math::

            \alpha(u) = \begin{cases}{
                    \frac{1}{\sqrt{2}}} &{\text{if }}u=0 \\
                    1 &{\text{otherwise}}
                \end{cases}

        There is technically no restriction on the range of pixel values but to match JPEG it is recommended to use the range [-128, 127].
    """
    N = blocks.shape[3]

    n = _normalize(N).to(blocks.device)
    h = _harmonics(N).to(blocks.device)

    coeff = (2 / N) * n * (h.t() @ blocks @ h)

    return coeff


def block_idct(coeff: Tensor) -> Tensor:
    r"""
    Computes the inverse DCT of non-overlapping blocks

    Args:
        coeff (Tensor): The blockwise DCT coefficients in the format :math:`(N, C, L, H, W)`

    Returns:
        Tensor: The pixels for each block in the same format as the input.

    Note:
        This function computes the inverse DCT given by

        .. math::
            I_{x,y}={\frac {1}{\sqrt{2N}}}\sum _{i=0}^{N}\sum _{j=0}^{N}\alpha (i)\alpha (j)D_{i,j}\cos \left[{\frac {(2x+1)i\pi }{2N}}\right]\cos \left[{\frac {(2y+1)j\pi }{2N}}\right]

        See :py:func:`block_dct` for further details.
    """
    N = coeff.shape[3]

    n = _normalize(N).to(coeff.device)
    h = _harmonics(N).to(coeff.device)

    im = (2 / N) * (h @ (n * coeff) @ h.t())
    return im


def batch_dct(batch: Tensor) -> Tensor:
    r"""
    Computes the DCT of a batch of images. See :py:func:`block_dct` for more details.
    This function takes care of splitting the images into blocks for the :py:func:`block_dct` and reconstructing
    the original shape of the input after the DCT.

    Args:
        batch (Tensor): A batch of images of format :math:`(N, C, H, W)`.

    Returns:
        Tensor: A batch of DCT coefficients of the same format as the input.

    Note:
        This fuction uses a block size of 8 to match the JPEG algorithm.
    """
    size = (batch.shape[2], batch.shape[3])

    im_blocks = blockify(batch, 8)
    dct_blocks = block_dct(im_blocks)
    dct = deblockify(dct_blocks, size)

    return dct


def batch_idct(coeff: Tensor) -> Tensor:
    r"""
    Computes the inverse DCT of a batch of coefficients. See :py:func:`block_dct` for more details.
    This function takes care of splitting the images into blocks for the :py:func:`block_idct` and reconstructing
    the original shape of the input after the inverse DCT.

    Args:
        batch (Tensor): A batch of coefficients of format :math:`(N, C, H, W)`.

    Returns:
        Tensor: A batch of images of the same format as the input.

    Note:
        This function uses a block size of 8 to match the JPEG algorithm.
    """
    size = (coeff.shape[2], coeff.shape[3])

    dct_blocks = blockify(coeff, 8)
    im_blocks = block_idct(dct_blocks)
    im = deblockify(im_blocks, size)

    return im


def fdct(im: Tensor) -> Tensor:
    r"""
    Convenience function for taking the DCT of a single image

    Args:
        im (Tensor): A single image of format :math:`(C, H, W)`

    Returns:
        Tensor: The DCT coefficients of the input in the same format.

    Note:
        This function simply expands the input in the batch dimension and then calls :py:func:`batch_dct` then removes
        the added batch dimension of the result.
    """
    return batch_dct(im.unsqueeze(0)).squeeze(0)


def idct(coeff: Tensor) -> Tensor:
    r"""
    Convenience function for taking the inverse InversDCT of a single image

    Args:
        im (Tensor): DCT coefficients of format :math:`(C, H, W)`

    Returns:
        Tensor: The image pixels of the input in the same format.

    Note:
        This function simply expands the input in the batch dimension and then calls :py:func:`batch_idct` then removes
        the added batch dimension of the result.
    """
    return batch_idct(coeff.unsqueeze(0)).squeeze(0)


def normalize(dct: Tensor, stats: Stats, channel: Optional[str] = None) -> Tensor:
    r"""
    Normalizes DCT coefficients using pre-computed stats

    This function wraps the :py:class:`DCTStats` class to allow easy normalization of multichannel images

    Args:
        dct (Tensor): The DCT coefficients to normalize in the format :math:`(N, C, H, W)`
        stats (:py:class:`DCTStats`): Precomputed DCT statistics
        channel (Optional[str]): If the input coefficients are single channel, the Y channel is assumed, use this parameter to override that.

    Returns:
        Normalized DCT coefficients in the format :math:`(N, C, H, W)`
    """
    ch = []

    for i in range(dct.shape[1]):
        dct_blocks = blockify(dct[:, i : (i + 1), :, :], 8)

        t = ["y", "cb", "cr"][i] if channel is None else channel
        dct_blocks = stats.normalize(dct_blocks, normtype=t)

        ch.append(deblockify(dct_blocks, (dct.shape[2], dct.shape[3])))

    return torch.cat(ch, dim=1)


def denormalize(dct: Tensor, stats: Stats, channel: Optional[str] = None) -> Tensor:
    r"""
    Denormalizes DCT coefficients using pre-computed stats

    This function wraps the :py:class:`DCTStats` class to allow easy denormalization of multichannel images

    Args:
        dct (Tensor): The DCT normalized coefficients in the format :math:`(N, C, H, W)`
        stats (:py:class:`DCTStats`): Precomputed DCT statistics
        channel (Optional[str]): If the input coefficients are single channel, the Y channel is assumed, use this parameter to override that.

    Returns:
        Denormalized DCT coefficients in the format :math:`(N, C, H, W)`
    """
    ch = []

    for i in range(dct.shape[1]):
        dct_blocks = blockify(dct[:, i : (i + 1), :, :], 8)

        t = ["y", "cb", "cr"][i] if channel is None else channel
        dct_blocks = stats.denormalize(dct_blocks, normtype=t)

        ch.append(deblockify(dct_blocks, (dct.shape[2], dct.shape[3])))

    return torch.cat(ch, dim=1)


def batch_to_images(dct: Tensor, stats: Optional[Stats] = None, crop: Optional[Tensor] = None, channel: Optional[str] = None) -> Tensor:
    r"""
    Converts a batch of DCT coefficients to a batch of images.

    This high level convenience function wraps several operations. If stats are given, the coefficients are
    assumed to have been channel-wise and frequency-wise normalized and are denormalized. The coefficients are tranformed to pixels and uncentered (converted from
    [-128, 127] to [0, 255]). If the input is multichannel, it is converted from YCbCr to RGB. The image is then optionally cropped to remove padding that may
    have been added to make the coefficients. The output is then rescaled to [0, 1] to match pytorch conventions.

    Args:
        dct (Tensor): A batch of DCT coefficients in :math:`(N, C, H, W)` format.
        stats (Optional[:py:class:DCTStats]): Optional DCT per-channel and per-frequency statistics to denormalize the coefficients.
        crop (Optional[Tensor]): Optional cropping dimensions. If this tensor has more than a single dimension, only the last dimension is used.
        channel (Optional[str]): One of 'Y', 'Cb', 'Cr'. Denormalization of a single channel input assumes 'Y' channel by default, use this paramter to override that.

    Returns:
        Tensor: The batch of images computed from the given coefficients.

    """
    if stats is not None:
        dct = denormalize(dct, stats, channel=channel)

    spatial = batch_idct(dct) + 128

    if spatial.shape[1] == 3:
        spatial = to_rgb(spatial)

    spatial = spatial.clamp(0, 255)
    spatial = spatial / 255

    if crop is not None:
        while len(crop.shape) > 1:
            crop = crop[0]

        cropY = crop[-2]
        cropX = crop[-1]

        spatial = spatial[:, :, : int(cropY), : int(cropX)]

    return spatial


def images_to_batch(spatial: Tensor, stats: Optional[Stats] = None, channel: Optional[str] = None) -> Tensor:
    r"""
    Converts a batch of images to a batch of DCT coefficients.

    This high level convenience function wraps several operations. The input images are assumed to follow the pytorch convention of being in [0, 1]
    and are rescaled to [0, 255]. If the images are multichannel, they are converted to YCbCr and then centered (in [-128, 127]). The
    DCT is taken and if stats are given, the coefficients are normalized.

    Args:
        spatial (Tensor): A batch of images in :math:`(N, C, H, W)` format.
        stats (Optional[:py:class:DCTStats]): Optional DCT per-channel and per-frequency statistics to normalize the coefficients.
        channel (Optional[str]): One of 'Y', 'Cb', 'Cr'. Normalization of a single channel input assumes 'Y' channel by default, use this parameter to override that.


    Returns:
        A batch of DCT coefficients computed from the input images.
    """
    spatial *= 255

    if spatial.shape[1] == 3:
        spatial = to_ycbcr(spatial)

    spatial -= 128

    frequency = batch_dct(spatial)
    if stats is not None:
        return normalize(frequency, stats, channel=channel)

    return frequency


def pad_to_block_multiple(im: Tensor, macroblock_size: int = 16) -> Tensor:
    r"""
    Pads an image to make it fit an even number of blocks

    Args:
        im (Tensor): the input image to pad, can be of shape :math:`(N, C, H, W)` or :math:`(C, H, W)`.
        macroblock_size (int): The size of a macroblock, sometimes referred to as a minimum coded unit (MCU), default 16.

    Note:
        As in the JPEG standard, the padding is applied to the right and bottom edges and is replicate padding. Note that the default
        macroblock is of size 16 (meaning a :math:`16 \times 16` block). This ensures that if chroma subsampling is used on the color
        channels, after half-sizing they will still fit an even number of :math:`8 \times 8` blocks as required by the JPEG DCT.
    """
    s = torch.tensor(list(im.shape))

    if len(s) != 4:
        im = im.unsqueeze(0)

    p = (torch.ceil(s / macroblock_size) * macroblock_size - s).long()
    im = pad(im, [0, int(p[-1].item()), 0, int(p[-2].item())], "replicate")

    if len(s) != 4:
        im = im.squeeze(0)

    return im


def zigzag(coefficients: Tensor) -> Tensor:
    r"""
    Vectorizes a DCT coefficients in JPEG zigzag order

    Args:
        coefficients (Tensor): DCT coefficients of shape :math:`(N, C, H, W)` or :math:`(C, H, W)`.

    Returns:
        Tensor: A batch of vectorized coefficients of shape :math:`(N, C, L, 64)` or :math:`(C, L, 64)`

    Note:
        For a visual representation of JPEG zigzag order see https://en.wikipedia.org/wiki/JPEG#/media/File:JPEG_ZigZag.svg.
    """
    assert len(coefficients.shape) in (3, 4)

    # Don't mess with matrix formatting
    # fmt: off
    zigzag_indices = torch.tensor([
         0,  1,  5,  6, 14, 15, 27, 28,
         2,  4,  7, 13, 16, 26, 29, 42,
         3,  8, 12, 17, 25, 30, 41, 43,
         9, 11, 18, 24, 31, 40, 44, 53,
        10, 19, 23, 32, 39, 45, 52, 54,
        20, 22, 33, 38, 46, 51, 55, 60,
        21, 34, 37, 47, 50, 56, 59, 61,
        35, 36, 48, 49, 57, 58, 62, 63
    ]).long()
    # fmt: on

    if len(coefficients.shape) == 3:
        c = coefficients.unsqueeze(0)
    else:
        c = coefficients

    c = blockify(c, 8)

    c = c.view(c.shape[0], c.shape[1], c.shape[2], 64)
    c = c[..., zigzag_indices]

    if len(coefficients.shape) == 3:
        c = c.squeeze(0)

    return c


__all__ = [
    "blockify",
    "deblockify",
    "block_dct",
    "block_idct",
    "batch_dct",
    "batch_idct",
    "fdct",
    "idct",
    "to_ycbcr",
    "to_rgb",
    "normalize",
    "denormalize",
    "batch_to_images",
    "images_to_batch",
    "double_nn_dct",
    "half_nn_dct",
    "Stats",
    "pad_to_block_multiple",
    "zigzag",
]



r"""
The :py:mod:`torchjpeg.quantization` package provides functions which quantize DCT coefficients. The IJG (libjpeg) quantization matrices
are included as part of this package as well as code which generates them from a scalar quality factor. Users can also provide their 
own quantization matrices. This implementation of the IJG quantization matrices only deals with the "baseline" setting where the 
maximum quantization value is 255. Functions in this module operate on single channel images since the channels are often quantized  
separately or may not be at the same resolution. 
"""


from typing import Callable, Optional, Tuple

import torch
from torch import Tensor
from torch.nn.functional import upsample

from torchjpeg.dct import blockify, deblockify, double_nn_dct, half_nn_dct

from . import ijg
from ._quantize import dequantize, quantize


def quantize_multichannel(dct: Tensor, mat: Tensor, round_func: Callable[[Tensor], Tensor] = torch.round) -> Tuple[Tensor, Tensor, Tensor]:
    r"""
    Quantizes a three channel image of DCT coefficients.

    Args:
        dct (Tensor): DCT coefficients of shape :math:`(N, 3, H, W)`.
        mat (Tensor): Quantization matrix of shape :math:`(1, 2, 8, 8)`.
        round: (Callable[[Tensor], Tensor]): Rounding function to use, defaults to :py:func:`torch.round`.

    Returns
    -------
        Tensor
            Y channel coefficients of shape :math:`(N, 1, H, W)`.
        Tensor
            Cb channel coefficients of shape :math:`\left(N, 1, \frac{H}{2}, \frac{W}{2}\right)`.
        Tensor
            Cr channel coefficients of shape :math:`\left(N, 1, \frac{H}{2}, \frac{W}{2}\right)`.

    Note:
        This function performs chroma subsampling
    """
    y_coefficients = dct[:, 0:1, :, :]
    cb_coefficients = dct[:, 1:2, :, :]
    cr_coefficients = dct[:, 2:3, :, :]

    cb_coefficients = half_nn_dct(cb_coefficients)
    cr_coefficients = half_nn_dct(cr_coefficients)

    y_coefficients = quantize(
        y_coefficients,
        mat[:, 0:1, :, :],
        round_func=round_func,
    )
    cb_coefficients = quantize(cb_coefficients, mat[:, 1:2, :, :], round_func=round_func)
    cr_coefficients = quantize(cr_coefficients, mat[:, 1:2, :, :], round_func=round_func)

    return y_coefficients, cb_coefficients, cr_coefficients


def dequantize_multichannel(y: Tensor, cb: Tensor, cr: Tensor, mat: Tensor) -> Tensor:
    r"""
    Dequantizes a three channel image.

    Args:
        y (Tensor): Quantized Y channel coefficients of shape :math:`(N, 1, H, W)`.
        cb (Tensor): Quantized  Cb channel coefficients of shape :math:`\left(N, 1, \frac{H}{2}, \frac{W}{2}\right)`.
        cr (Tensor): Quantized Cr channel coefficients of shape :math:`\left(N, 1, \frac{H}{2}, \frac{W}{2}\right)`.

    Returns:
        Tensor: A three channel image of DCT coefficients.

    Note:
        This function assumes chroma subsampling.
    """
    y_coefficients = dequantize(y, mat[:, 0:1, :, :])
    cb_coefficients = dequantize(cb, mat[:, 1:2, :, :])
    cr_coefficients = dequantize(cr, mat[:, 1:2, :, :])

    cb_coefficients = double_nn_dct(cb_coefficients)
    cr_coefficients = double_nn_dct(cr_coefficients)

    coefficients = torch.cat([y_coefficients, cb_coefficients, cr_coefficients], dim=1)
    return coefficients


__all__ = ["quantize", "dequantize", "quantize_multichannel", "dequantize_multichannel"]



from io import BytesIO
from random import randint
from typing import Tuple

from PIL import Image


class YChannel:
    r"""
    Converts a tensor with a color image in [0, 1] to the Y channel using ITU-R BT.601 conversion

    Warning:
        This is **not** equivalent to the Y channel of a color image that would be used by JPEG, the result
        is in [16, 240] following the ITU-R BT.601 standard before normalization. This is useful for certian JPEG artifact correction
        algorithms due to some questionable evaluation choices by that community. The result **is** normalized to :math:`\left[\frac{16}{255},\frac{240}{255}\right]`
        before being returned.
    """

    def __init__(self):
        pass

    def __call__(self, tensor):
        if tensor.shape[0] == 3:
            tensor = tensor * 255
            tensor = 16.0 + tensor[0, :, :] * 65.481 / 255.0 + tensor[1, :, :] * 128.553 / 255.0 + tensor[2, :, :] * 24.966 / 255.0
            tensor = tensor.unsqueeze(0).round() / 255.0

        return tensor

    def __repr__(self):
        return self.__class__.__name__ + "()"


class YCbCr:
    r"""
    Converts a PIL image to YCbCr color space

    Note:
        PIL follows the JPEG YCbCr color conversion giving a result in [0, 255].
    """

    def __init__(self):
        pass

    def __call__(self, pil_image):
        return pil_image.convert("YCbCr")

    def __repr__(self):
        return self.__class__.__name__ + "()"


class RandomJPEG:
    r"""
    Applies JPEG compression on a PIL at a random quality.

    Args:
        quality_range (Tuple[int, int]): The quality range to choose from, inclusive on both ends.
            An integer in this range will be chosen at random and will be used as the compression quality setting.
    """

    def __init__(self, quality_range: Tuple[int, int] = (0, 100)) -> None:
        self.quality_range = quality_range

    def __call__(self, pil_image):
        q = randint(*self.quality_range)
        with BytesIO() as f:
            pil_image.save(f, format="jpeg", quality=q)
            f.seek(0)
            output = Image.open(f)
            output.load()

        return output

    def __repr__(self):
        return f"{self.__class__.__name__} ({self.quality_range})"



r"""
Provides access to low-level JPEG operations using libjpeg.
By using libjpeg directly, coefficients can be loaded or saved to JPEG files directly with needing to be recomputed.
In addtion to the C++ implemented low-level operations, two python convenience functions are exported that can decode the ressulting coefficients to pixels.
"""
from typing import Optional

import torch
from torch import Tensor
from torch.nn.functional import interpolate

from torchjpeg.dct import block_idct, deblockify, double_nn_dct, to_rgb

from ._codec_ops import *  # pylint: disable=import-error

__all__ = ["read_coefficients", "write_coefficients", "quantize_at_quality", "pixels_for_channel", "reconstruct_full_image"]  # pylint: disable=undefined-all-variable


def pixels_for_channel(channel: Tensor, quantization: Tensor, crop: Optional[Tensor] = None) -> Tensor:
    r"""
    Converts a single channel of quantized DCT coefficients into pixels.

    Args
    ----------
    channel : torch.Tensor
        A :math:`\left(1, \frac{H}{8}, \frac{W}{8}, 8, 8 \right)` Tensor of quantized DCT coefficients.
    quantization : torch.Tensor
        An (8, 8) Tensor of the quantization matrix that was used to quantize :code:`channel`.
    crop : torch.Tensor
        An optional (2) Tensor of containing the `$\left(H, W \right)$` original sizes of the image channel stored in :code:`channel`. The pixel result will be cropped to this size.

    Returns
    -------
    torch.Tensor
        A :math:`\left(H, W \right)` Tensor containing the pixel values of the channel in [0, 1]

    Note
    -----
    This function takes inputs in the same format as returned by :py:func:`read_coefficients` separated into a single channel.
    """
    dequantized = channel.float() * quantization.float()

    s = block_idct(dequantized) + 128
    s = s.view(1, 1, s.shape[1] * s.shape[2], 8, 8)
    s = deblockify(s, (channel.shape[1] * 8, channel.shape[2] * 8))
    s = s.squeeze()

    if crop is not None:
        s = s[: int(crop[0]), : int(crop[1])]

    return s


def reconstruct_full_image(y_coefficients: Tensor, quantization: Tensor, cbcr_coefficients: Optional[Tensor] = None, crop: Optional[Tensor] = None) -> Tensor:
    r"""
    Converts quantized DCT coefficients into an image.

    This function is designed to work on the output of :py:func:`read_coefficients` and py:func:`quantize_at_quality`. Note that the color channel coefficients
    will be upsampled by 2 as chroma subsampling is currently assumed. If the image is color, it will be converted from YCbCr to RGB.

    Parameters
    ----------
    y_coefficients : torch.Tensor
        A :math:`\left(1, \frac{H}{8}, \frac{W}{8}, 8, 8 \right)` Tensor of quantized Y channel DCT coefficients.
    quantization : torch.Tensor
        A :math:`\left(C, 8, 8 \right)` Tensor of quantization matrices for each channel.
    cbcr_coefficients : Optional[torch.Tensor]
        A :math:`\left(2, \frac{H}{8}, \frac{W}{8}, 8, 8 \right)` Tensor of quantized color channel DCT coeffcients.
    crop : Optional[torch.Tensor]
        A :math:`\left(C, 2 \right)` Tensor containing the :math:`\left(H, W \right)` dimensions of the image that produced the given  DCT coefficients, the pixel result will be cropped to this size.

    Returns
    -------
    torch.Tensor
        A :math:`\left(C, H, W \right)` Tensor containing the image pixels in pytorch format (normalized to [0, 1])
    """
    y = pixels_for_channel(y_coefficients, quantization[0], crop[0] if crop is not None else None)

    if cbcr_coefficients is not None:
        cb = pixels_for_channel(cbcr_coefficients[0:1], quantization[1], crop[1] if crop is not None else None)
        cr = pixels_for_channel(cbcr_coefficients[1:2], quantization[2], crop[2] if crop is not None else None)

        cb = interpolate(cb.unsqueeze(0).unsqueeze(0), y.shape, mode="nearest")
        cr = interpolate(cr.unsqueeze(0).unsqueeze(0), y.shape, mode="nearest")

        out = torch.cat([y.unsqueeze(0).unsqueeze(0), cb, cr], dim=1)
        out = to_rgb(out).squeeze()
    else:
        out = y

    return out.clamp(0, 255) / 255.0



from typing import Tuple

import torch
import torch.nn.functional
from torch import Tensor


def blockify(im: Tensor, size: int) -> Tensor:
    r"""
    Breaks an image into non-overlapping blocks of equal size.

    Parameters
    ----------
    im : Tensor
        The image to break into blocks, must be in :math:`(N, C, H, W)` format.
    size : Tuple[int, int]
        The size of the blocks in :math:`(H, W)` format.

    Returns
    -------
    A tensor containing the non-overlappng blocks in :math:`(N, C, L, H, W)` format where :math:`L` is the
    number of non-overlapping blocks in the image channel indexed by :math:`(N, C)` and :math:`(H, W)` matches
    the block size.

    Note
    ----
    If the image does not split evenly into blocks of the given size, the result will have some overlap. It
    is the callers responsibility to pad the input to a multiple of the block size, no error will be thrown
    in this case.
    """
    bs = im.shape[0]
    ch = im.shape[1]
    h = im.shape[2]
    w = im.shape[3]

    im = im.reshape(bs * ch, 1, h, w)
    im = torch.nn.functional.unfold(im, kernel_size=(size, size), stride=(size, size))
    im = im.transpose(1, 2)
    im = im.reshape(bs, ch, -1, size, size)

    return im


def deblockify(blocks: Tensor, size: Tuple[int, int]) -> Tensor:
    r"""
    Reconstructs an image given non-overlapping blocks of equal size.

    Args:
        blocks (Tensor): The non-overlapping blocks in :math:`(N, C, L, H, W)` format.
        size: (Tuple[int, int]): The dimensions of the original image (e.g. the desired output)
            in :math:`(H, W)` format.

    Returns:
        The image in :math:`(N, C, H, W)` format.

    Note:
        If the blocks have some overlap, or if the output size cannot be constructed from the given number of non-overlapping
        blocks, this function will raise an exception unlike :py:func:`blockify`.

    """
    bs = blocks.shape[0]
    ch = blocks.shape[1]
    block_size = blocks.shape[3]

    blocks = blocks.reshape(bs * ch, -1, int(block_size**2))
    blocks = blocks.transpose(1, 2)
    blocks = torch.nn.functional.fold(blocks, output_size=size, kernel_size=(block_size, block_size), stride=(block_size, block_size))
    blocks = blocks.reshape(bs, ch, size[0], size[1])

    return blocks



from typing import Callable

import torch
from torch import Tensor

from torchjpeg.dct import blockify, deblockify


def quantize(dct: Tensor, mat: Tensor, round_func: Callable[[Tensor], Tensor] = torch.round) -> Tensor:
    r"""
    Quantizes DCT coefficients.

    Args:
        dct (Tensor): DCT coefficients of shape :math:`(N, 1, H, W)`.
        mat: (Tensor): Quantization matrix of shape :math:`(1, 1, 8, 8)`.
        round_func: (Callable[[Tensor], Tensor]): Rounding function to use, defaults to :py:func:`torch.round`.

    Returns:
        Tensor: Quantized DCT coefficients.

    Note:
        DCT quantization is computed as

        .. math::
            \widetilde{D}_{ij} = \left\lfloor \frac{D_{ij}}{Q_{ij}} \right\rceil

        For DCT coefficients :math:`D` and quantization matrix :math:`Q`.
    """
    dct_blocks = blockify(dct, 8)
    quantized_blocks = round_func(dct_blocks / mat)
    quantized = deblockify(quantized_blocks, (dct.shape[2], dct.shape[3]))
    return quantized


def dequantize(dct: Tensor, mat: Tensor) -> Tensor:
    r"""
    Dequantize DCT coefficients.

    Args:
        dct (Tensor): Quantized DCT coefficients of shape :math:`(N, 1, H, W)`.
        mat: (Tensor): Quantization matrix of shape :math:`(1, 1, 8, 8)`.

    Returns:
        Tensor: Quantized DCT coefficients.

    Note:
        DCT dequantization is computed as

        .. math::
            D_{ij} = \widetilde{D}_{ij} \cdot Q_{ij}

        For quantized DCT coefficients :math:`\widetilde{D}` and quantization matrix :math:`Q`.
    """
    dct_blocks = blockify(dct, 8)
    dequantized_blocks = dct_blocks * mat
    dequantized = deblockify(dequantized_blocks, (dct.shape[2], dct.shape[3]))
    return dequantized



import math
from typing import Optional

import torch
from torch import Tensor

from ._block import blockify, deblockify


def double_size_tensor() -> Tensor:
    """box resize, takes and 8 x 8 image and returns a 16 x 16 image"""
    op = torch.zeros((8, 8, 16, 16))
    for i in range(0, 8):
        for j in range(0, 8):
            for u in range(0, 16):
                for v in range(0, 16):
                    if i == u // 2 and j == v // 2:
                        op[i, j, u, v] = 1

    return op


def half_size_tensor() -> Tensor:
    """box resize, takes a 16 x 16 and returns an 8 x 8"""
    op = torch.zeros((16, 16, 8, 8))
    for i in range(0, 16):
        for j in range(0, 16):
            for u in range(0, 8):
                for v in range(0, 8):
                    if i == 2 * u and j == 2 * v:
                        op[i, j, u, v] = 1

    return op


def A(alpha) -> float:
    """DCT orthonormal scale factor"""
    if alpha == 0:
        return 1.0 / math.sqrt(2)

    return 1.0


def D() -> Tensor:
    """DCT tensor"""
    D_t = torch.zeros((8, 8, 8, 8))

    for i in range(8):
        for j in range(8):
            for alpha in range(8):
                for beta in range(8):
                    scale_a = A(alpha)
                    scale_b = A(beta)

                    coeff_x = math.cos(((2 * i + 1) * alpha * math.pi) / 16)
                    coeff_y = math.cos(((2 * j + 1) * beta * math.pi) / 16)

                    D_t[i, j, alpha, beta] = 0.25 * scale_a * scale_b * coeff_x * coeff_y
    return D_t


def reblock() -> Tensor:
    """reblocker, takes a 16 x 16 and returns 4 8 x 8 blocks"""
    B_t = torch.zeros((16, 16, 4, 8, 8))

    # This is deep but it really needs to be
    # pylint: disable=too-many-nested-blocks
    for s_x in range(16):
        for s_y in range(16):
            for n in range(4):
                for i in range(8):
                    for j in range(8):
                        x = n % 2
                        y = n // 2
                        if x * 8 + i == s_x and y * 8 + j == s_y:
                            B_t[s_x, s_y, n, i, j] = 1.0

    return B_t


def macroblock() -> Tensor:
    """Takes 4 x 8 x 8 and returns a 16 x 16"""
    B_t = torch.zeros((4, 8, 8, 16, 16))

    # 0 goes in top left
    for alpha in range(8):
        for beta in range(8):
            B_t[0, alpha, beta, alpha, beta] = 1

    # 1 goes in top right
    for alpha in range(8):
        for beta in range(8):
            B_t[1, alpha, beta, alpha + 8, beta] = 1

    # 2 goes in bottom left
    for alpha in range(8):
        for beta in range(8):
            B_t[2, alpha, beta, alpha, beta + 8] = 1

    # 3 goes in bottom right
    for alpha in range(8):
        for beta in range(8):
            B_t[3, alpha, beta, alpha + 8, beta + 8] = 1

    return B_t


class ResizeOps:  # pylint: disable=too-few-public-methods
    """
    Caches the resize operation tensors and enables them to be built on demand. Generally not for public use
    """

    resizer: Optional[Tensor] = None
    halfsizer: Optional[Tensor] = None
    dct: Optional[Tensor] = None
    reblocker: Optional[Tensor] = None
    macroblocker: Optional[Tensor] = None
    block_doubler: Optional[Tensor] = None
    block_halver: Optional[Tensor] = None

    @classmethod
    def lazy_build_ops(cls) -> None:
        """
        Builds the resize operations
        """

        # HACK assume none of the operators are set if the first one isnt set
        if cls.resizer is None:
            cls.resizer = double_size_tensor()

            cls.halfsizer = half_size_tensor()
            cls.dct = D()
            cls.reblocker = reblock()
            cls.macroblocker = macroblock()

            # block doubler combines the following linear operations in order: inverse DCT, NN doubling, reshape to 4 x 8 x 8, DCT, reshape back to 16 x 16
            cls.block_doubler = torch.einsum("ijab,ijmn,mnzxy,xypq,zpqrw->abrw", cls.dct, cls.resizer, cls.reblocker, cls.dct, cls.macroblocker)
            # 16 x 16 -> 4 x 8 x 8 -> idct -> 16 x 16 -> resize -> dct
            cls.block_halver = torch.einsum("mnzab,ijab,zijrw,rwxy,xypq->mnpq", cls.reblocker, cls.dct, cls.macroblocker, cls.halfsizer, cls.dct)


def double_nn_dct(input_dct: Tensor, op: Optional[Tensor] = None) -> Tensor:
    r"""
    double_nn_dct(input_dct: Tensor, op: Tensor = block_doubler) -> Tensor:

    DCT domain nearest neighbor doubling

    The function computes a 2x nearest neighbor upsampling on DCT coefficients without converting them to pixels.
    It is equivalent to the following procedure: IDCT -> 2x upsampling -> DCT

    Args:
        input_dct (Tensor): The input DCT coefficients in the format :math:`(N, C, H, W)`
        op (Tensor): The doubling operation tensor, mostly used to satisfy torchscript. Should be of shape :math:`8 \times 8 \times 16 \times 16`. Leave as default unless you know what you're doing.

    Returns:
        Tensor: The coefficients of the resized image, double the height and width of the input.
    """
    if op is None:
        ResizeOps.lazy_build_ops()
        op = ResizeOps.block_doubler

    if op is not None:
        op = op.to(input_dct.device)

    dct_blocks = blockify(input_dct, 8)
    dct_doubled = torch.einsum("abrw,ncdab->ncdrw", [op, dct_blocks])
    deblocked_doubled = deblockify(dct_doubled, (input_dct.shape[2] * 2, input_dct.shape[3] * 2))

    return deblocked_doubled


def half_nn_dct(input_dct: Tensor, op: Optional[Tensor] = None) -> Tensor:
    r"""
    half_nn_dct(input_dct: Tensor, op: Tensor = block_halver) -> Tensor:

    DCT domain nearest neighbor half-sizing

    The function computes a 2x nearest neighbor downsampling on DCT coefficients without converting them to pixels.
    It is equivalent to the following procedure: IDCT -> 2x downsampling -> DCT

    Args:
        input_dct (Tensor): The input DCT coefficients in the format :math:`(N, C, H, W)`
        op (Tensor): The halving operation tensor, mostly used to satisfy torchscript. Should be of shape :math:`16 \times 16 \times 8 \times 8`. Leave as default unless you know what you're doing.

    Returns:
        Tensor: The coefficients of the resized image, halg the height and width of the input.
    """
    if op is None:
        ResizeOps.lazy_build_ops()
        op = ResizeOps.block_halver

    if op is not None:
        op = op.to(input_dct.device)

    dct_blocks = blockify(input_dct, 16)
    dct_halved = torch.einsum("abrw,ncdab->ncdrw", [op, dct_blocks])
    deblocked_halved = deblockify(dct_halved, (input_dct.shape[2] // 2, input_dct.shape[3] // 2))

    return deblocked_halved



r"""
The :code:`torchjpeg.metrcs` package provides useful metrics for measuring JPEG quality. All the 
metrics in this package take inputs in format :math:`(N, C, H, W)` and produces outputs of format
:math:`(N)` by averaging spatially and over channels. The batch dimension is not averaged. Inputs
should be images in [0, 1].
"""
from ._psnr import *
from ._psnrb import *
from ._size import *
from ._ssim import *

__all__ = ["psnr", "psnrb", "blocking_effect_factor", "ssim", "size"]



import torch
from torch import Tensor


def psnr(image: Tensor, target: Tensor) -> Tensor:
    r"""
    Computes the peak signal-to-noise ratio on a batch of images.

    Args:
        image (Tensor): Input images in format :math:`(N, C, H, W)`.
        target (Tensor): Target images in format :math:`(N, C, H, W)`.

    Returns:
        Tensor: PSNR for each image in the batch, of shape :math:`(N)`.

    Note:
        Peak signal-to-noise ratio is an inverse log scale of the mean squared error
        measured in decibels. The formula used here is

        .. math::

            P(x, y) = 10 \log_{10}\left(\frac{1}{\text{MSE}(x, y)}\right)

    """
    mse = torch.nn.functional.mse_loss(image, target, reduction="none")
    mse = mse.view(mse.shape[0], -1).mean(1)
    return 10 * torch.log10(1 / mse)



from pathlib import Path
from typing import Any, Callable, Optional, Sequence, Union

import torch
from PIL import Image


class UnlabeledImageFolder(torch.utils.data.Dataset):
    """
    Dataset loading a folder of unlabeled images recursively. The images are loaded using PIL and otherwise unchanged, add a transform to turn them into Tensors

    Args:
        path (Path): The path to load recursively from
        extensions (List[str]): The image extensions to look for
        transform: Any transform to apply to the images after loading them
    """

    def __init__(self, path: Union[str, Path], extensions: Sequence[str] = [".bmp", ".png", ".jpg", ".ppm", ".pgm"], transform: Optional[Callable] = None) -> None:
        # pylint: disable=dangerous-default-value
        if isinstance(path, str):
            path = Path(path)

        self.path = path
        self.images = list(filter(lambda p: p.suffix in extensions, path.glob("**/*")))
        self.transform = transform

    def __len__(self) -> int:
        return len(self.images)

    def __getitem__(self, idx) -> Any:
        im_path = self.images[idx]

        with open(im_path, "rb") as f:
            im = Image.open(f)
            im.load()

        if self.transform is not None:
            im = self.transform(im)

        return im



import importlib.resources
from pathlib import Path
from typing import Union

import torch
from torch import Tensor


class Stats:
    r"""
    This class holds pre-computed per-channel and per-frequency DCT coefficient stats.

    The stats are loaded from a file, this can be written using :py:func:`torch.save`.
    The file should contain a single dictionary with string keys containing channel names.
    The value of each entry should be a dictionary with the keys: "mean, variance, min, and max"
    with the corresponding statistics as Tensors.

    Pre-computed stats are available for color or grayscale images (pass "color" and "grayscale" respectively for
    the root argument), these stats were computed from the Flickr 2k dataset, a large corpus of high quality images and
    are suitable for general use.

    Args:
        root (:py:class:`pathlib.Path`, string, or literals "color", "grayscale"): The path to load the statistics from or "color" to use built in color stats or "grayscale" to use built in grayscale stats.
        normtype (str): Either "ms" for mean-variance normalization or "01" for zero-one normalization.
    """

    def __init__(self, root: Union[str, Path], normtype: str = "ms") -> None:
        self.type = normtype

        if root in ("color", "grayscale"):
            reader = importlib.resources.open_binary("torchjpeg.dct.stats", f"{root}.pt")
        else:
            if isinstance(root, str):
                root = Path(root)

            reader = root.open("rb")

        stats = torch.load(reader)

        self.mean = {x: stats[x]["mean"].view(1, 1, 8, 8) for x in stats.keys()}
        self.variance = {x: stats[x]["variance"].view(1, 1, 8, 8) for x in stats.keys()}
        self.std = {x: torch.sqrt(self.variance[x]) for x in stats.keys()}

        self.min = {x: stats[x]["min"].view(1, 1, 8, 8) for x in stats.keys()}
        self.max = {x: stats[x]["max"].view(1, 1, 8, 8) for x in stats.keys()}

    def normalize(self, blocks: Tensor, normtype: str = "y") -> Tensor:
        r"""
        Normalizes blocks of coefficients.

        Args:
            blocks (Tensor): a Tensor containing blocks of DCT coefficients in the format :math:`(N, C, L, H, W)`.
            normtype (str): Which channel to normalize, "y" by default.

        Returns:
            Tensor: The normalized coefficients.
        """
        if self.type == "ms":
            return self._mean_variance_f(blocks, normtype)
        if self.type == "01":
            return self._zero_one_f(blocks, normtype)

        raise NotImplementedError(f"Unknown norm type {normtype}, must be 01 or ms")

    def denormalize(self, blocks: Tensor, normtype: str = "y") -> Tensor:
        r"""
        Denormalizes blocks of coefficients.

        Args:
            blocks (Tensor): a Tensor containing blocks of normalized DCT coefficients in the format :math:`(N, C, L, H, W)`.
            normtype (str): Which channel to denormalize, "y" by default.

        Returns:
            Tensor: The denormalized coefficients.
        """
        if self.type == "ms":
            return self._mean_variance_r(blocks, normtype)
        if self.type == "01":
            return self._zero_one_r(blocks, normtype)

        raise NotImplementedError(f"Unknown norm type {normtype}, must be 01 or ms")

    def _mean_variance_f(self, blocks: Tensor, normtype: str = "y") -> Tensor:
        m = self.mean[normtype].to(blocks.device)
        s = self.std[normtype].to(blocks.device)

        return (blocks - m) / s

    def _zero_one_f(self, blocks: Tensor, normtype: str = "y") -> Tensor:
        m = -self.min[normtype].to(blocks.device)
        s = self.max[normtype] - self.min[normtype]
        s = s.to(blocks.device)

        return (blocks + m) / s

    def _mean_variance_r(self, blocks: Tensor, normtype: str = "y") -> Tensor:
        s = self.std[normtype].to(blocks.device)
        m = self.mean[normtype].to(blocks.device)

        return blocks * s + m

    def _zero_one_r(self, blocks: Tensor, normtype: str = "y") -> Tensor:
        s = self.max[normtype] - self.min[normtype]
        s = s.to(blocks.device)
        m = -self.min[normtype].to(blocks.device)

        return blocks * s - m



from .folder_of_jpeg_dataset import FolderOfJpegDataset
from .image_list import ImageList, crop_batch
from .jpeg_quantized_dataset import JPEGQuantizedDataset
from .unlabeled_image_folder import UnlabeledImageFolder

__all__ = ["ImageList", "crop_batch", "JPEGQuantizedDataset", "FolderOfJpegDataset", "UnlabeledImageFolder"]



from io import BytesIO
from typing import Tuple, Union

import torch
from PIL import Image
from torch import Tensor
from torchvision.transforms.functional import to_pil_image, to_tensor


def size(image: Union[Tensor, Image.Image], **kwargs) -> Tuple[Tensor, Tensor]:
    r"""
    Computes the size in bytes of a JPEG

    Args:
        image (Tensor or PIL Image): The image to compress
        kwargs: Arguments to pass to the PIL JPEG compressor (like quality or quantization matrices)

    Returns
    -------
        Tensor
            A single element tensor containing the size in bytes of the image after JPEG compression
        Tensor
            The compressed image

    Warning:
        The output of this function is **not** differentiable. It compresses the image to memory and reads the size of
        the resulting buffer.
    """
    if isinstance(image, Tensor):
        image = to_pil_image(image)

    with BytesIO() as f:
        image.save(f, "jpeg", **kwargs)
        f.seek(0)
        s = f.getbuffer().nbytes

        im = Image.open(f)
        im.load()
        im = to_tensor(im)

    return torch.tensor([s]), im



from random import randrange
from typing import Any, Mapping, Optional, Sequence, Tuple, Union

import torch
from torch.functional import Tensor
from torch.utils.data import Dataset

from torchjpeg.codec import quantize_at_quality  # type: ignore # pylint: disable=no-name-in-module
from torchjpeg.dct import Stats, deblockify, images_to_batch, normalize, pad_to_block_multiple
from torchjpeg.quantization.ijg import quantization_max

from .image_list import ImageList


def _dequantize_channel(channel, quantization):
    dequantized_dct = channel.float() * quantization
    dequantized_dct = dequantized_dct.view(1, 1, dequantized_dct.shape[1] * dequantized_dct.shape[2], 8, 8)
    dequantized_dct = deblockify(dequantized_dct, (channel.shape[1] * 8, channel.shape[2] * 8))

    return dequantized_dct


def _prep_coefficients(quantization: Tensor, Y_coefficients: Tensor, CbCr_coefficients: Tensor, stats: Stats):
    quantization = quantization.float()
    y_q = quantization[0]

    y_dequantized = _dequantize_channel(Y_coefficients, y_q)
    y_q = y_q / quantization_max
    y_dequantized = normalize(y_dequantized, stats, channel="y")

    if CbCr_coefficients is not None:
        c_q = quantization[1]  # Assume same quantization for cb and cr

        cb_dequantized = _dequantize_channel(CbCr_coefficients[0:1], c_q)
        cr_dequantized = _dequantize_channel(CbCr_coefficients[1:2], c_q)

        c_q = c_q / quantization_max

        cb_dequantized = normalize(cb_dequantized, stats, channel="cb")
        cr_dequantized = normalize(cr_dequantized, stats, channel="cr")

        cbcr_dequantized = torch.cat([cb_dequantized, cr_dequantized], dim=1)
    else:
        cbcr_dequantized = torch.empty(0)
        c_q = torch.empty(0)

    return y_dequantized, cbcr_dequantized, y_q, c_q


class JPEGQuantizedDataset(Dataset):
    r"""
    Wraps an arbitrary image dataset to return JPEG quantized versions of the image. The amount quantization is defined using IJG quality settings.
    If the underlying dataset returns a sequence, the first element of the sequence is taken the be the image which is quantized and the remaining
    elements are returned as the last element of the batch. If the underlying dataset returns a mapping, set `image_key` to the key of the image
    to be quantized. The original dictionary, including the image before quantization, will be returned as the last element of the batch.

    Since the primary return values are all DCT coefficients, padding will be added to the images to make them an even multiple of the MCU. Following JPEG
    conventions this is replicate padding added to the bottom and right edges. The original size of the image is returned so that the images can be
    correctly cropped after processing.

    The format returned by this dataset is:

    Y Channel Coefficients, CbCr Coefficients, Y Quantization Matrix, CbCr Quantization Matrix, Pre-quantization YCbCr Coefficients, Original Image Size, Optional rest of the batch from the underlying dataset

    If the image is grayscale, the CbCr coefficients and quantization matrix will be an empty tensor, if the underlying dataset returns an image with no additional data, the final return value will be an empty tensor.
    This is to avoid issues with the default collate function, an empty tensor is one initialized with 0 size using :py:func:`torch.empty`. It is detectable as `tensor.numel() == 0`.

    Args:
        data (:py:class:`torch.utils.data.Dataset`): The dataset to wrap
        quality (int, tuple of two or three ints): The quality range (min 0 max 100) to draw from, inclusive on both ends. If this is a single integer, only that quality is used, if it's three integers, the last one defines a step size.
        stats (:py:clas:`torchjpeg.dct.Stats`): Statstics to use for per-frequency per-channel coefficient normalization
        mcu (int): The size of the minimum coded unit, use 16 for 4:2:0 chroma subsampling.
        image_key (optional str): The key to use to extract the image from a dataset which returns a mapping.
        deterministic_quality (bool): False by default, set to True to include the quality range in the dataset size. In other words, the length of this dataset will be `len(quality_range) * len(dataset)` and all the qualities in the range will
            be represented for every image by interating this dataset.

    Warning:
        The images returned from this dataset may be of differing sizes, use the static :py:func:`torchjpeg.data.JPEGQuantizedDataset.collate` to collate them into a batch with padding. Use :py:func:`torchjpeg.data.crop_batch` to
        crop them back to the correct sizes (this will also remove JPEG padding).
    """

    def __init__(self, dataset: Dataset, quality: Union[int, Sequence[int]], stats: Stats, mcu: int = 16, image_key: Optional[str] = None, deterministic_quality: bool = False) -> None:
        # pylint: disable=too-many-arguments
        assert (isinstance(quality, Sequence) and len(quality) > 0) or isinstance(quality, int)

        if isinstance(quality, int):
            quality = (quality, quality + 1, 1)
        elif isinstance(quality, Sequence):
            if len(quality) == 1:
                quality = (quality[0], quality[0] + 1, 1)
            elif len(quality) == 2:
                quality = (quality[0], quality[1] + 1, 1)
            elif len(quality) > 2:
                quality = (quality[0], quality[1], quality[2])

        self.dataset = dataset
        self.quality = quality
        self.stats = stats
        self.mcu = mcu
        self.image_key = image_key
        self.deterministic_quality = deterministic_quality

    def __len__(self) -> int:
        if self.deterministic_quality:
            return len(self.dataset) * len(range(*self.quality))  # type: ignore

        return len(self.dataset)  # type: ignore

    def __getitem__(self, idx) -> Tuple[Tensor, Tensor, Tensor, Tensor, Tensor, Tensor, Any]:
        # pylint: disable=too-many-locals
        if not self.deterministic_quality:
            image = self.dataset[idx]
            quality = randrange(*self.quality)
        else:
            image_idx = idx % len(self.dataset)  # type: ignore
            quality_idx = idx // len(self.dataset)  # type: ignore

            image = self.dataset[image_idx]
            quality = quality_idx * self.quality[2] + self.quality[0]

        labels: Any = torch.empty(0)

        if isinstance(image, Sequence):
            labels = image[1:]
            image = image[0]
        elif isinstance(image, Mapping):
            labels = image
            image = image[self.image_key]

        s = torch.as_tensor(image.shape)
        image = pad_to_block_multiple(image)

        (
            _,
            quantization,
            Y_coefficients,
            CbCr_coefficients,
        ) = quantize_at_quality(image, quality)

        groundtruth_dct = images_to_batch(image.unsqueeze(0), self.stats).squeeze(0)  # This *has* to be after quantization

        y_dequantized, cbcr_dequantized, y_q, c_q = _prep_coefficients(quantization, Y_coefficients, CbCr_coefficients, self.stats)

        return (
            y_dequantized.squeeze(0),
            cbcr_dequantized.squeeze(0),
            y_q.unsqueeze(0),
            c_q.unsqueeze(0),
            groundtruth_dct,
            s.long(),
            labels,
        )

    @staticmethod
    def collate(batch_list: Sequence[Tuple[Tensor, Tensor, Tensor, Tensor, Tensor, Tensor, Any]]) -> Tuple[Tensor, Tensor, Tensor, Tensor, Tensor, Tensor, Any]:
        """
        Custom collate function which works for return values from this dataset. Adds padding to the images so that they can be stored in a single tensor

        Args:
            batch_list: Output from this dataset

        Returns:
            Batch with each input collated into single tensors
        """
        y_coefs = []
        cbcr_coefs = []
        gt_coefs = []

        yqs = torch.stack([b[2] for b in batch_list])
        cqs = torch.stack([b[3] for b in batch_list])
        sizes = torch.stack([b[5] for b in batch_list])

        labels = [b[6] for b in batch_list]

        for b in batch_list:
            y_coefs.append(b[0])
            cbcr_coefs.append(b[1])
            gt_coefs.append(b[4])

        y_coefs_t = ImageList.from_tensors(y_coefs).tensor
        cbcr_coefs_t = ImageList.from_tensors(cbcr_coefs).tensor
        gt_coefs_t = ImageList.from_tensors(gt_coefs).tensor

        return y_coefs_t, cbcr_coefs_t, yqs, cqs, gt_coefs_t, sizes, labels



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['torchjpeg',
 'torchjpeg.codec',
 'torchjpeg.data',
 'torchjpeg.data.transforms',
 'torchjpeg.dct',
 'torchjpeg.dct.stats',
 'torchjpeg.metrics',
 'torchjpeg.quantization']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=7.2.0', 'torch>=2.0.0,<2.0.1', 'torchvision>=0.15.1,<0.15.2']

setup_kwargs = {
    'name': 'torchjpeg',
    'version': '0.9.32',
    'description': 'Utilities for JPEG data access and manipulation in pytorch',
    'long_description': '# TorchJPEG\n\n[![pipeline status](https://gitlab.com/torchjpeg/torchjpeg/badges/master/pipeline.svg)](https://gitlab.com/torchjpeg/torchjpeg/-/commits/master) \n[![coverage report](https://gitlab.com/torchjpeg/torchjpeg/badges/master/coverage.svg)](https://gitlab.com/torchjpeg/torchjpeg/-/commits/master) \n[![PyPI](https://img.shields.io/pypi/v/torchjpeg)](https://pypi.org/project/torchjpeg/)\n[![License](https://img.shields.io/badge/license-MIT-blue)](https://gitlab.com/Queuecumber/torchjpeg/-/blob/master/LICENSE)\n\nThis package contains a C++ extension for pytorch that interfaces with libjpeg to allow for manipulation of low-level JPEG data.\nBy using libjpeg, quantization results are guaranteed to be consistent with other applications, like image viewers or MATLAB,\nwhich use libjpeg to compress and decompress images. This is useful because JPEG images can be effected by round-off\nerrors or slight differences in the decompression procedure. Besides this, this library can be used to read and write\nDCT coefficients, functionality which is not available from other python interfaces.\n\nBesides this, the library includes many utilities related to JPEG compression, many of which are written using native pytorch code meaning\nthey can be differentiated or GPU accelerated. The library currently includes packages related to the DCT, quantization, metrics, and dataset\ntransformations.\n\n## LIBJPEG\n\nCurrently builds against: `libjpeg-9d`. libjpeg is statically linked during the build process. See [http://www.ijg.org/files/](http://www.ijg.org/files/) for libjpeg source. \nThe full libjpeg source is included with the torchjpeg source code and will be built during the install process (for a source or sdist install).\n\n## Install\n\nPackages are hosted on [pypi](https://pypi.org/project/torchjpeg/). Install using pip, note that only Linux builds are supported at the moment. \n\n```\npip install torchjpeg\n```\n\nIf there is demand for builds on other platforms it may happen in the future. Also note that the wheel is intended to be compatible with manylinux2014\nwhich means it should work on modern Linux systems, however, because of they way pytorch works, we can\'t actually build it using all of the manylinux2014\ntools. So compliance is not guaranteed and YMMV.\n\n```{warning}\ntorchjpeg is currently in pre-beta development and consists mostly of converted research code. The public facing API, including any and all names of\nparameters and functions, is subject to change at any time. We follow semver for versioning and will adhere to that before making and breaking\nchanges.\n```\n\n## Citation\n\nIf you use our code in a publication, we ask that you cite the following paper ([bibtex](http://softmax.me/bibtex/ehrlich2020quantization.txt)):\n\n> Max Ehrlich, Larry Davis, Ser-Nam Lim, and Abhinav Shrivastava. "Quantization Guided JPEG Artifact Correction." In Proceedings of the European Conference on Computer Vision, 2020',
    'author': 'Max Ehrlich',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://torchjpeg.readthedocs.io',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



import torch
from torch import Tensor


def to_ycbcr(x: Tensor, data_range: float = 255) -> Tensor:
    r"""
    Converts a Tensor from RGB color space to YCbCr color space

    Parameters
    ----------
    x : Tensor
        The input Tensor holding an RGB image in :math:`(\ldots, C, H ,W)` format (where :math:`\ldots` indicates an arbitrary number of dimensions).
    data_range : float
        The range of the input/output data. i.e., 255 indicates pixels in [0, 255], 1.0 indicates pixels in [0, 1]. Only 1.0 and 255 are supported.

    Returns
    -------
    Tensor
        The YCbCr result of the same shape as the input and with the same data range.

    Note
    -----
    This function implements the "full range" conversion used by JPEG, e.g. it does **not** implement the ITU-R BT.601 standard which
    many libraries (excluding PIL) use as the default definition of YCbCr. This conversion (for [0, 255]) is given by:

    .. math::
        \begin{aligned}
        Y&=&0&+(0.299&\cdot R)&+(0.587&\cdot G)&+(0.114&\cdot B) \\
        C_{B}&=&128&-(0.168736&\cdot R)&-(0.331264&\cdot G)&+(0.5&\cdot B) \\
        C_{R}&=&128&+(0.5&\cdot R)&-(0.418688&\cdot G)&-(0.081312&\cdot B)
        \end{aligned}

    """
    assert data_range in [1.0, 255]

    # fmt: off
    ycbcr_from_rgb = torch.tensor([
        0.29900, 0.58700, 0.11400,
        -0.168735892, -0.331264108, 0.50000,
        0.50000, -0.418687589, -0.081312411
    ],
    device=x.device).view(3, 3).transpose(0, 1)
    # fmt: on

    if data_range == 255:
        b = torch.tensor([0, 128, 128], device=x.device).view(3, 1, 1)
    else:
        b = torch.tensor([0, 0.5, 0.5], device=x.device).view(3, 1, 1)

    x = torch.einsum("cv,...cxy->...vxy", [ycbcr_from_rgb, x])
    x += b

    return x.contiguous()


def to_rgb(x: Tensor, data_range: float = 255) -> Tensor:
    r"""
    Converts a Tensor from YCbCr color space to RGB color space

    Parameters
    ----------
    x : Tensor
        The input Tensor holding a YCbCr image in :math:`(\ldots, C, H ,W)` format (where :math:`\ldots` indicates an arbitrary number of dimensions).
    data_range : float
        The range of the input/output data. i.e., 255 indicates pixels in [0, 255], 1.0 indicates pixels in [0, 1]. Only 1.0 and 255 are supported.

    Returns
    -------
    Tensor
        The RGB result of the same shape as the input and with the same data range.

    Note
    -----
    This function expects the input to be "full range" conversion used by JPEG, e.g. it does **not** implement the ITU-R BT.601 standard which
    many libraries (excluding PIL) use as the default definition of YCbCr. If the input came from this library or from PIL it should be fine.
    The conversion (for [0, 255]) is given by:

    .. math::
        \begin{aligned}
        R&=&Y&&&+1.402&\cdot (C_{R}-128) \\
        G&=&Y&-0.344136&\cdot (C_{B}-128)&-0.714136&\cdot (C_{R}-128 ) \\
        B&=&Y&+1.772&\cdot (C_{B}-128)&
        \end{aligned}

    """
    assert data_range in [1.0, 255]

    # fmt: off
    rgb_from_ycbcr = torch.tensor([
        1, 0, 1.40200,
        1, -0.344136286, -0.714136286,
        1, 1.77200, 0
    ],
    device=x.device).view(3, 3).transpose(0, 1)
    # fmt: on

    if data_range == 255:
        b = torch.tensor([-179.456, 135.458816, -226.816], device=x.device).view(3, 1, 1)
    else:
        b = torch.tensor([-0.70374902, 0.531211043, -0.88947451], device=x.device).view(3, 1, 1)

    x = torch.einsum("cv,...cxy->...vxy", [rgb_from_ycbcr, x])
    x += b

    return x.contiguous()



import numpy as np
import torch
from torch import Tensor


def blocking_effect_factor(im: Tensor) -> Tensor:
    r"""
    Computes the blocking effect factor (BEF) of an image as defined in [1].

    Blocking effect factor is used as part of :py:func:`psnrb` but can also be used
    as an objective measure of "blockiness".

    Args:
        im (Tensor): Image of shape :math:`(N, C, H, W)`.

    Returns:
        Tensor: The BEF for each image of shape :math:`(N)`.

    Note:
        [1] Tadala, Trinadh, and Sri E. Venkata Narayana. "A Novel PSNR-B Approach for Evaluating the Quality of De-blocked Images." (2012).
    """
    # This is a fairly complex formula we're implementing, it would be messy with fewer locals
    # pylint: disable=too-many-locals
    block_size = 8

    block_horizontal_positions = torch.arange(7, im.shape[3] - 1, 8)
    block_vertical_positions = torch.arange(7, im.shape[2] - 1, 8)

    horizontal_block_difference = ((im[:, :, :, block_horizontal_positions] - im[:, :, :, block_horizontal_positions + 1]) ** 2).sum(3).sum(2).sum(1)
    vertical_block_difference = ((im[:, :, block_vertical_positions, :] - im[:, :, block_vertical_positions + 1, :]) ** 2).sum(3).sum(2).sum(1)

    nonblock_horizontal_positions = np.setdiff1d(torch.arange(0, im.shape[3] - 1), block_horizontal_positions)
    nonblock_vertical_positions = np.setdiff1d(torch.arange(0, im.shape[2] - 1), block_vertical_positions)

    horizontal_nonblock_difference = ((im[:, :, :, nonblock_horizontal_positions] - im[:, :, :, nonblock_horizontal_positions + 1]) ** 2).sum(3).sum(2).sum(1)
    vertical_nonblock_difference = ((im[:, :, nonblock_vertical_positions, :] - im[:, :, nonblock_vertical_positions + 1, :]) ** 2).sum(3).sum(2).sum(1)

    n_boundary_horiz = im.shape[2] * (im.shape[3] // block_size - 1)
    n_boundary_vert = im.shape[3] * (im.shape[2] // block_size - 1)
    boundary_difference = (horizontal_block_difference + vertical_block_difference) / (n_boundary_horiz + n_boundary_vert)

    n_nonboundary_horiz = im.shape[2] * (im.shape[3] - 1) - n_boundary_horiz
    n_nonboundary_vert = im.shape[3] * (im.shape[2] - 1) - n_boundary_vert
    nonboundary_difference = (horizontal_nonblock_difference + vertical_nonblock_difference) / (n_nonboundary_horiz + n_nonboundary_vert)

    scaler = np.log2(block_size) / np.log2(min([im.shape[2], im.shape[3]]))
    bef = scaler * (boundary_difference - nonboundary_difference)

    bef[boundary_difference <= nonboundary_difference] = 0
    return bef


def psnrb(image: Tensor, target: Tensor) -> Tensor:
    r"""
    Computes the peak signal-to-noise ratio with blocking effect factor from [1].

    PSNR-B augments the PSNR measure by including the "blockiness" of the degraded image as a way to reduce
    the PSNR. For multichannel inputs, the PSNR-B is computed separately for each channel and then averaged.

    Args:
        image (Tensor): The input images of shape :math:`(N, C, H, W)`.
        target (Tensor): The target images of shape :math:`(N, C, H, W)`.

    Returns:
        Tensor: The PSNR-B of each image in the batch of shape :math:`(N)`.

    Warning:
        Unlike most metrics this is not symmetric and the order of the arguents is imporant. Blocking effect factor
        is only computed for the degraded image, so if the arguments are reversed, there will be very little difference
        between this and :py:func:`psnr`.

    Note:
        PSNR-B is computed as

        .. math::
            P(x, y) = 10 \log_{10}\left(\frac{1}{\text{MSE}(x, y) + \text{BEF}(x)}\right)

        [1] Tadala, Trinadh, and Sri E. Venkata Narayana. "A Novel PSNR-B Approach for Evaluating the Quality of De-blocked Images." (2012).
    """

    def channel_psnrb(image, target):
        mse = torch.nn.functional.mse_loss(image, target, reduction="none")
        bef = blocking_effect_factor(image)

        mse = mse.view(mse.shape[0], -1).mean(1)
        return 10 * torch.log10(1 / (mse + bef))

    total = torch.stack([channel_psnrb(image[:, c : c + 1, ...], target[:, c : c + 1, ...]) for c in range(image.shape[1])]).sum(0)
    return total / image.shape[1]
